package drifter;



import net.sourceforge.jFuzzyLogic.FIS;
import net.sourceforge.jFuzzyLogic.rule.FuzzyRuleSet;
import net.sourceforge.jFuzzyLogic.rule.Variable;
import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.mg4j.util.MutableString;
import simplerace.*;


public class MyFuzzyController implements Constants, Controller {
	FuzzyRuleSet fuzzyRuleSet;
	FuzzyRuleSet speedRuleSet;
	Variable heading_angle,steer,speed,distance,next_angle,heading,next,dist,expected_speed,next_distance;
	final static int maxSteps=150;
	final static int minDistanceToDrift=150;
	final static double maxDistance=3000;
	final static double maxAngle = 3.14159;
	private double diagonal = Math.sqrt(400*400 + 400*400);
	public MyFuzzyController() {
		// TODO Auto-generated constructor stub
		String fileName = "fcl/steering.fcl";//name of the system
		FIS fis = FIS.load(fileName, false);//load the system
		if( fis == null )
		{ // Error while loading?
			System.err.println("Can't load file: '" + fileName + "'");
			return;
		}//load the system from here
		fuzzyRuleSet = fis.getFuzzyRuleSet();//create a fuzzy rule set from the system
		
		// TODO Auto-generated constructor stub
		fileName = "fcl/speed.fcl";//name of the system
		fis = FIS.load(fileName, false);//load the system
		if( fis == null )
		{ // Error while loading?
			System.err.println("Can't load file: '" + fileName + "'");
			return;
		}//load the system from here
		speedRuleSet = fis.getFuzzyRuleSet();//create a fuzzy rule set from the system

		// Show ruleset
		
		heading_angle = fuzzyRuleSet.getVariable(new MutableString("heading_angle"));
		distance = fuzzyRuleSet.getVariable(new MutableString("distance"));
		next_angle = fuzzyRuleSet.getVariable(new MutableString("next_angle"));
		steer = fuzzyRuleSet.getVariable(new MutableString("steer"));
		
//		speed = speedRuleSet.getVariable(new MutableString("speed"));
		heading = speedRuleSet.getVariable(new MutableString("heading_angle"));
		dist = speedRuleSet.getVariable(new MutableString("distance"));
		next = speedRuleSet.getVariable(new MutableString("next_angle"));
		speed = speedRuleSet.getVariable(new MutableString("speed"));
//		next_distance = speedRuleSet.getVariable(new MutableString("next_distance"));
		expected_speed = speedRuleSet.getVariable(new MutableString("expected_speed"));
//		fuzzyRuleSet.chart();
//		Vector2d A = new Vector2d(0,0);
//		Vector2d B = new Vector2d(1,0);
//		Vector2d C = new Vector2d(2,0);
//		System.out.println(angle(A, B, C));



		
	}
	
	public static double angle(Vector2d A,Vector2d B,Vector2d C){
		double xDiff = C.x - B.x;
        double yDiff = C.y - B.y;
        double angle = Math.atan2(yDiff, xDiff);
        xDiff = A.x - B.x;
        yDiff = A.y - B.y;
        return Utils.correctAngle(angle-Math.atan2(yDiff, xDiff));
	}
	
	public static double getAngleToWaypoint(Vector2d pos, double orientation, Vector2d wp)
    {        
        double xDiff = wp.x - pos.x;
        double yDiff = wp.y - pos.y;
        double angle = Math.atan2(yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle(angle);
    }
    
	public static double correctAngle (double angle) {
        while (angle < - Math.PI)
            angle += Math.PI * 2;
        while (angle > Math.PI)
            angle -= Math.PI * 2;
        return angle;
    }
	
	public static double getDistance(Vehicle car,Vector2d wp){
		return (car.getPosition().dist(wp));
	}
	
	public static double getHeadingAngle(Vehicle car,Vector2d wp){
		Vehicle tmp = car.copy();
		double dist = getDistance(car, wp);
		
		double d=0;
		while (d<dist){
			tmp.update(forward);			
			d = getDistance(car, tmp.getPosition());
		}
						
		return angle(tmp.getPosition(),car.getPosition(),wp);
	}
	
	public static double getSpeedToNextWaypoint(Vector2d velocity, double orientation)
    {
    	double velocityMagnitude = velocity.mag ();
        double directionOfMovement = Math.atan2 (velocity.y, velocity.x);
        // rotate
        double directionDifference = directionOfMovement - orientation;        
        return Math.cos (directionDifference) * velocityMagnitude;
    }
	public static double getSpeed(Vehicle car){
    	return getSpeedToNextWaypoint(car.getVelocity(), car.getOrientation());
    }
	
	public Turning path(Vehicle car,Vector2d wp,Vector2d nwp){
		IntArrayList p = new IntArrayList();
		DoubleArrayList d = new DoubleArrayList();
		ObjectArrayList<Vehicle> steps = new ObjectArrayList<Vehicle>();
		
		Vehicle tmp = car.copy();
		Turning rs = new Turning(tmp,wp,maxSteps);
		
		int count=0;
		boolean ok =(getDistance(tmp, wp)<=waypointIntersectionRadius);
		while (!ok && count++<=maxSteps){
			int cmd = gotoTarget(tmp, wp, nwp);
			tmp.update(cmd);
			p.add(cmd);
			double dist = getDistance(tmp, wp); 
			d.add(dist);
			steps.add(tmp.copy());
			ok = (dist<=waypointIntersectionRadius);
		}
		if (!ok) return null;
		rs.setDist(d);
		rs.setSteps(steps);
		rs.setCommands(p);
		rs.setNoOfSteps(p.size());		
		return rs;
	}
	int minFound=0;
	public Turning drift(Vehicle car,Vector2d wp,Vector2d nwp,int maxSteps){
		IntArrayList p = new IntArrayList();
		DoubleArrayList d = new DoubleArrayList();
		ObjectArrayList<Vehicle> steps = new ObjectArrayList<Vehicle>();
		
		Vehicle tmp = car.copy();
		Turning rs = new Turning(tmp,nwp,maxSteps);
		minFound=0;
		int count=0;
		double oldDist = getDistance(tmp, wp);
		boolean ok =(oldDist<=waypointIntersectionRadius);
		while (count++<=maxSteps){			
			int cmd = gotoTarget(tmp, nwp, nwp);
			tmp.update(cmd);
			p.add(cmd);
			double dist = getDistance(tmp, wp); 
			d.add(dist);
			steps.add(tmp.copy());
			if (!ok) {
				++minFound;
				ok = (dist<=waypointIntersectionRadius);
			}
			
			if (!ok && oldDist<dist) break;
			oldDist = dist;
		}
		if (!ok) return null;
		rs.setDist(d);
		rs.setSteps(steps);
		rs.setCommands(p);
		rs.setNoOfSteps(p.size());		
		return rs;
	}
	
	public static boolean isTheSame(Vector2d v1,Vector2d v2){
    	return (v1.x==v2.x && v1.y==v2.y);
    }
	
	public static Turning sensible(Vehicle car,Vector2d wp,boolean forw,int maxStep){
		Vehicle tmp = car.copy();		
		double orientation = tmp.getOrientation();
		
		double distance = tmp.getPosition() .dist(wp);
		double d = distance;
		double theta =  Math.atan((waypointIntersectionRadius)/distance)/2;			
		double heading = correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));		
		Turning t = new Turning(tmp,wp,maxSteps);
		ObjectArrayList<Vehicle> steps = t.getSteps();
		DoubleArrayList dist = t.getDist();
		IntArrayList cmds = t.getCommands();		
		int i = 0;
		int cmd =0;
//		boolean ok = v.dist(new Vector2d(-6.130760991518019 ,0))==0;
		
		while (d>=0 && distance>waypointIntersectionRadius && i<maxStep){
			if (!forw) 
				cmd = (Math.abs(heading)<=theta) ? backward : (heading>0) ? backwardleft : backwardright;
			else cmd = (Math.abs(heading)<=theta) ? forward :(heading>0) ? forwardleft : forwardright;
			cmds.add(cmd);
						
			tmp.update(cmd);
			d-=getSpeed(tmp);
			++i;			
			orientation = tmp.getOrientation();
			distance = tmp.getPosition() .dist(wp);
			theta = Math.atan((waypointIntersectionRadius)/distance)/2;	
			heading = (forw) ?correctAngle(getAngleToWaypoint(tmp.getPosition(), orientation, wp)):
				correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));
			dist.add(distance);
			steps.add(tmp.copy());
		}
		
		
//		if (i==maxSteps && Math.abs(heading)>theta/2 && distance>waypointIntersectionRadius) return null;
//		if (ok) System.out.println(i);		
		t.setNoOfSteps(i);
		return t;
	}
	
	public Turning flatOut(Vehicle car,Vector2d wp,Vector2d nwp,int minStep){
		double speed = getSpeed(car);		
//		Turning p = (speed<=0 || Math.abs(heading)>Math.PI/2-0.4 || nextDistance<60) ?path(car,wp,nwp):straightOnly(car, wp);
		
		Turning p = (speed<=0) ? sensible(car, wp, false, minStep+1) :sensible(car, wp, true, minStep+1);
		if (p!=null && p.getCommands().size()>minStep) return null;
//		System.out.println(p.getNoOfSteps());
		if (p==null) return null;
		if (isTheSame(wp, nwp)  && Turning.isHitTarget(p.getSteps(),wp)!=null) return p;		
		
		ObjectArrayList<Vehicle> steps = p.getSteps();
		DoubleArrayList dist = p.getDist();
		int count=0;
		double min=30000;
		boolean found = false;
		Turning rs = null;
		
		for (int i=0;i<p.getNoOfSteps();++i){
			double d = dist.getDouble(i);
			if (d>minDistanceToDrift) continue;
			Vehicle tmp = steps.get(i);
			Turning pp = drift(tmp,wp,nwp,maxSteps);
//			if (pp!=null && min<pp.getNoOfSteps()+i+5)
//				break;
			if (pp!=null && min>pp.getNoOfSteps()+i && minFound<minStep){
				found=true;
				count++;
				min = pp.getNoOfSteps()+i;
				rs = Turning.combine(p, i, pp, wp);
			}
			if (found) count++;
			if (count>5 ) break;
		}
//		System.out.println(count);
		
		return rs;
	}
	
	public Turning evaluate(Vehicle car,Vector2d wp,Vector2d nwp,int minStep){
		double speed = getSpeed(car);
		
//		Turning p = (speed<=0 || Math.abs(heading)>Math.PI/2-0.4 || nextDistance<60) ?path(car,wp,nwp):straightOnly(car, wp);
		
		Turning p = path(car,wp,nwp);
		if (p!=null && p.getCommands().size()>minStep) return null;
//		System.out.println(p.getNoOfSteps());
		if (p==null) return null;
		if (isTheSame(wp, nwp)) return p;		
		
		ObjectArrayList<Vehicle> steps = p.getSteps();
		DoubleArrayList dist = p.getDist();
		int count=0;
		double min=30000;
		boolean found = false;
		Turning rs = null;
		
		for (int i=0;i<p.getNoOfSteps();++i){
			double d = dist.getDouble(i);
			if (d>minDistanceToDrift) continue;
			Vehicle tmp = steps.get(i);
			Turning pp = drift(tmp,wp,nwp,maxSteps);
//			if (pp!=null && min<pp.getNoOfSteps()+i+5)
//				break;
			if (pp!=null && min>pp.getNoOfSteps()+i && minFound<minStep){
				found=true;
				count++;
				min = pp.getNoOfSteps()+i;
				rs = Turning.combine(p, i, pp, wp);
			}
			if (found) count++;
			if (count>5 ) break;
		}		
		
		return rs;
	}
	
	
//	public boolean isTooClose(Vehicle car,Vector2d wp,double r){
//		double heading = Math.min(maxAngle,Math.max(getHeadingAngle(car, wp),-maxAngle));
//		Vector2d pos = car.getPosition();
//		Vector2d velocity = car.getVelocity();
//		double orientation = car.getOrientation();		
//        double directionOfMovement = Math.atan2 (velocity.y, velocity.x);
//        double speed = getSpeed(car);
//		double slipAngle = directionOfMovement - (speed >= 0 ? orientation : orientation + Math.PI);
//		double orthogonalDirection = (heading < 0 ? -(Math.PI / 2) : (Math.PI / 2));
//        Vector2d centre = new Vector2d
//                (Math.cos (orientation + orthogonalDirection),
//                        Math.sin (orientation + orthogonalDirection));
//        centre.setMag(r);
//        centre.add(pos);
//        System.out.println(pos+"    "+heading+"    "+centre);
//		return centre.dist(wp)<=r;
//	}
//	 int count=0;
//	
	public static double correct(double angle,double maxAngle){
		return Math.min(maxAngle,Math.max(angle,-maxAngle));
	}
	 
	public int gotoTarget(Vehicle car,Vector2d wp,Vector2d nwp){
		double heading = Math.min(maxAngle,Math.max(getHeadingAngle(car, wp),-maxAngle));
		double next =  Math.min(maxAngle,Math.max(getHeadingAngle(car, nwp),-maxAngle));
		double distance = Math.min(getDistance(car, wp),maxDistance);
		double nextDist = Math.min(wp.dist(nwp),maxDistance);
		double angle = Math.max(Math.abs(angle(car.getPosition(), wp, nwp)),maxAngle);
		double theta = Math.abs(correct(Math.atan(waypointIntersectionRadius/distance)/2,maxAngle));
		
		int s=0;
		double speed = getSpeed(car);		
		if ((speed>0 && distance>50) || (speed<=0 && distance>40) ){			
			this.heading.setValue(heading);
			this.next.setValue(next);
			dist.setValue(distance);
//			System.out.println(this.heading);
	//		next_distance.setValue(nextDist);		
			this.speed.setValue(speed);
			
			speedRuleSet.evaluate();
			
			
			double sp = expected_speed.getLatestDefuzzifiedValue();
			
			s =(sp>=speed )? 2 :(sp>0) ? 1:0;			
		} else {
			
			if (speed>0){
				s=(speed>4.5) ? 0 : (speed>6.5 && distance>40)?1: 
					(Math.abs(heading)<=2*theta) ? 2 : (Math.abs(heading)>=Math.PI/2-0.4 && distance<=40) ? 0 : (speed>2.5) ? 0 :2;
				
			} else if (speed<=0) {
				s=(speed<-6.5 && distance>40)?1:(Math.PI-Math.abs(heading)<=2*theta)?0:
					(Math.PI-Math.abs(heading)>Math.PI/2+0.4 && distance<=40)? 2 : (speed<-2.5) ? 2 :0;
//				s=0;
			} else s=0;
//			System.out.println(distance+"    "+(Math.abs(heading))+"    "+speed+"   "+s);
		}
		
//		System.out.println(s);
		
		if (s==0){
			heading = correctAngle(Math.PI - heading);
		}
		double n = 5,m=3;
		if (angle>Math.PI/8 && nextDist>83 && s>0){			
			theta = theta*2;
		}
		if (s==0) theta/=2;
		heading_angle.getMembershipFunction("StraightAhead").setParameter(0,correct(-2*theta,maxAngle));
		heading_angle.getMembershipFunction("StraightAhead").setParameter(1,correct(-theta,maxAngle));
		heading_angle.getMembershipFunction("StraightAhead").setParameter(2,correct(theta,maxAngle));
		heading_angle.getMembershipFunction("StraightAhead").setParameter(3,correct(2*theta,maxAngle));
		
		heading_angle.getMembershipFunction("SmallLeft").setParameter(0,correct(-n*theta,maxAngle));
		heading_angle.getMembershipFunction("SmallLeft").setParameter(1,correct(-m*theta,maxAngle));
		heading_angle.getMembershipFunction("SmallLeft").setParameter(2,correct(-2*theta,maxAngle));
		heading_angle.getMembershipFunction("SmallLeft").setParameter(3,correct(-theta,maxAngle));
		
		heading_angle.getMembershipFunction("SmallRight").setParameter(0,correct(theta,maxAngle));
		heading_angle.getMembershipFunction("SmallRight").setParameter(1,correct(2*theta,maxAngle));
		heading_angle.getMembershipFunction("SmallRight").setParameter(2,correct(m*theta,maxAngle));
		heading_angle.getMembershipFunction("SmallRight").setParameter(3,correct(n*theta,maxAngle));
		
		heading_angle.getMembershipFunction("LeftDir").setParameter(2,correct(n*theta,maxAngle));
		heading_angle.getMembershipFunction("RightDir").setParameter(1,correct(n*theta,maxAngle));
		
		heading_angle.setValue(heading);		
		this.distance.setValue(distance);
		next_angle.setValue(next);
		
		fuzzyRuleSet.evaluate();
		
		double fSteer = steer.getLatestDefuzzifiedValue();
//		System.out.println(fuzzyRuleSet);
		int cmd = (fSteer<0) ? 2 : (fSteer>0) ? 0:1;
//		int cmd = (s==0) ? (((Math.abs(heading)<=theta) || speed>-3) ? 1:(fSteer<0) ? 2 :  0) :
//			((fSteer<0) ? 2 :  (fSteer>0) ?0 :1) ;
		return cmd+s*3;		

	}
	
	Vector2d oldWp = null;
    IntArrayList stps=null,nexts=null;
    
    public boolean changeWp(Vector2d wp){
    	return (oldWp==null || oldWp.x != wp.x || oldWp.y != wp.y);    		
    }
    
    public int underDog(Vehicle car,Vector2d wp){
    	double distance = getDistance(car, wp);
    	if (distance>60) return gotoTarget(car, wp, wp);
    	double theta = Math.abs(correct(Math.atan(waypointIntersectionRadius/distance)/2,maxAngle));
    	double heading = getHeadingAngle(car, wp);
    	 int steering = heading < 0 ? 0 : (heading<=2*theta) ? 1 : 2;
         double limit = 7;
         
         if (distance < 0.2*diagonal)
             limit = 3;
         if (distance < 0.10*diagonal)
             limit = 1;
         if (distance < 0.040*diagonal)
             limit = 0;
         double speed = getSpeed (car);
         int driving = ( speed>=0) ?(speed > (2 * limit) ? 0 : (speed > limit ? 1 : 2)) :
         	(speed < (2 * -limit) ? 2 : (speed < -limit ? 1 : 0));
         return driving * 3 + steering;
    }
    
    public int expect(Vehicle car, Vector2d wp,Vector2d nwp,int num){
    	if (car==null) return 10000;
    	Vehicle tmp = car.copy();
    	for (int i=0;i<num;++i){
    		tmp.update(gotoTg(tmp, wp, nwp));
    		if (tmp.getPosition ().dist (wp) <= waypointIntersectionRadius)
    			return i;
    	}
    	return 1000;
    }
    
    public int gotoTg(Vehicle car, Vector2d wp,Vector2d nwp){
    	double distance = car.getPosition() .dist(wp);
    	if (distance<50) return gotoTarget(car, wp, nwp);
		double theta = Math.atan((waypointIntersectionRadius)/distance);
		double orientation = car.getOrientation();
		double speed = getSpeedToNextWaypoint(car.getVelocity(), car.getOrientation());
		double heading = getAngleToWaypoint(car.getPosition(), orientation, wp);		
		double u = speed;
		if (Math.abs(heading)<theta/2) {	
			double r = distance*distance-waypointIntersectionRadius*waypointIntersectionRadius;
    		r=(r<=0)?0:Math.sqrt(r);
    		int t = (int)((-u+Math.sqrt(u*u+2*0.4*r))/0.4)+1;
    		
    		double a = 2*(r-u*t)/t/t;
    		
//    		System.out.println(a);
			if (a>0 && (a/0.4)<=t) 
				return forward; 
			if (a==0) 
				return neutral;
			return backward;
		}
		
		if (heading>0) {
			return (heading>Math.PI/2) ? backwardleft : forwardleft;
		}
		if (heading<0) 
			return (heading<-Math.PI/2) ? backwardright : forwardright;
		return backward;
    }
    
    int loopCount=0;
	public int control(SensorModel inputs) {
		// TODO Auto-generated method stub
		Vehicle cur = inputs.getSimulation().getFirstVehicle();
		Vehicle other = inputs.getSimulation().getSecondVehicle();
		Vector2d wp = inputs.getNextWaypointPosition();
		Vector2d nwp = inputs.getNextNextWaypointPosition();
		if (cur.getPosition().dist(inputs.getPosition())!=0) {
			other = inputs.getSimulation().getFirstVehicle();
			cur = inputs.getSimulation().getSecondVehicle();			
		}
		
				
//		double hd =Math.abs(getHeadingAngle(cur, wp));
//		double nhd = Math.abs(getHeadingAngle(other, wp));
		
		int otherE = expect(other, wp, nwp, 150)+1;
		if (stps!=null && otherE<=minFound) {
			Turning t = flatOut(cur, wp, nwp,otherE);
			if (t==null){
				stps=null;
				return gotoTarget(cur, nwp, nwp);
			} 
			stps = t.getCommands();
		}
		if (stps!=null && stps.size()>0 && !changeWp(wp) && !cur.justCollided()) return stps.removeInt(0);
		
		int curE = expect(cur, wp, nwp, 150);
		double d = getDistance(cur, wp);
//		double otherDist = (other==null)?10000:getDistance(other, wp);
//		double speed=getSpeed(cur);
		if (curE>otherE && d>40){
		
//			double diff=d-otherDist;
			if ( loopCount<200){
				loopCount++;
				return underDog(cur, nwp);
			}
		} 
		
		double theta = Math.abs(correct(Math.atan(waypointIntersectionRadius/d)/2,maxAngle));
		if (d>minDistanceToDrift){
//			if (hd<=theta*2 && speed>0 && speed<=) return forward;
//			if (Math.PI-hd<=theta*2 && speed<0) return backward;
			return gotoTarget(cur, wp, nwp);
		}
//		
		if (stps==null || changeWp(wp) || cur.justCollided() ||  stps.isEmpty()){
			oldWp = wp;
			loopCount=0;
//			nexts = gotoTarget(cur, nwp, nwp,2,e2);
			Turning t = evaluate(cur, wp, nwp,otherE);
			if (t==null || t.getNoOfSteps()==0) return neutral;
			stps = t.getCommands();
			
		}
		return stps.removeInt(0);
//		return gotoTarget(cur, wp, nwp);
//		return underDog(cur, wp);
	}

	public void reset() {
		// TODO Auto-generated method stub

	}

}
