package drifter;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import simplerace.*;
public class Path implements Constants {
	public final static int num =100;
//	FuzzyRuleSet fuzzyRuleSet;
//	Variable heading_angle,steer,accel,sp;
	public final static int num_of_instantiations=1;
	final static double diagonal = Math.sqrt(400*400 + 400*400);
	final static double mass = 10;
    final static double constantDrag = 0.2;
    final static double motorForceForward = 4;
    final static double motorForceBackward = 2;
    final static double turningRadius = 25; // pixels. effective at low speeds
    final static double maxLateralTyreTraction = 2;
    final static double minimumSlipAngleToCareAbout = Math.PI / 16; // radians. for numerical stability 
    
    double initialSpeed,finalSpeed,expectedSpeed;
    double initialDist,finalDist;
    int maxSteps=1000;
    int noOfSteps;
    DoubleArrayList dist = new DoubleArrayList();
    DoubleArrayList steps = new DoubleArrayList();
    IntArrayList commands =  new IntArrayList();
    
	public Path(double initialSpeed, double expectedSpeed, double initialDist,
			int maxSteps, DoubleArrayList dist, DoubleArrayList steps,
			IntArrayList commands) {
		super();
		this.initialSpeed = initialSpeed;
		this.expectedSpeed = expectedSpeed;
		this.initialDist = initialDist;
		this.maxSteps = maxSteps;
		this.dist = dist;
		this.steps = steps;
		this.commands = commands;
	}
	
	public Path(double initialSpeed, double expectedSpeed, double initialDist,
			int maxSteps) {
		super();
		this.initialSpeed = initialSpeed;
		this.expectedSpeed = expectedSpeed;
		this.initialDist = initialDist;
		this.maxSteps = maxSteps;
	}
	
	

	public Path(double initialSpeed, double initialDist) {
		super();
		this.initialSpeed = initialSpeed;
		this.initialDist = initialDist;
	}

	public Path(double initialSpeed, double expectedSpeed, double initialDist) {
		super();
		this.initialSpeed = initialSpeed;
		this.expectedSpeed = expectedSpeed;
		this.initialDist = initialDist;
	}
	
	

	public int getNoOfSteps() {
		return noOfSteps;
	}

	public void setNoOfSteps(int noOfSteps) {
		this.noOfSteps = noOfSteps;
	}

	public double getInitialSpeed() {
		return initialSpeed;
	}

	public void setInitialSpeed(double initialSpeed) {
		this.initialSpeed = initialSpeed;
	}
	

	public double getFinalSpeed() {
		return finalSpeed;
	}
	
	public double getExpectedSpeed() {
		return expectedSpeed;
	}

	public void setExpectedSpeed(double expectedSpeed) {
		this.expectedSpeed = expectedSpeed;
	}

	public double getInitialDist() {
		return initialDist;
	}

	public void setInitialDist(double initialDist) {
		this.initialDist = initialDist;
	}

	public double getFinalDist() {
		return Math.abs(finalDist);
	}

	public int getMaxSteps() {
		return maxSteps;
	}

	public void setMaxSteps(int maxSteps) {
		this.maxSteps = maxSteps;
	}

	public DoubleArrayList getDist() {
		return dist;
	}

	public void setDist(DoubleArrayList dist) {
		this.dist = dist;
	}

	public DoubleArrayList getSteps() {
		return steps;
	}

	public void setSteps(DoubleArrayList steps) {
		this.steps = steps;
	}

	public IntArrayList getCommands() {
		return commands;
	}

	public void setCommands(IntArrayList commands) {
		this.commands = commands;
	}
	
	public static double updateSpeed(double speed,int command){
		if (command==forward)
			return speed+(motorForceForward-speed*constantDrag)/mass; 
		else if (command==backward)
			return speed-(motorForceBackward+speed*constantDrag)/mass;
		return speed-(speed*constantDrag/mass);  
	}
	
	
    
    public void setFinalSpeed(double finalSpeed) {
		this.finalSpeed = finalSpeed;
	}

	public void setFinalDist(double finalDist) {
		this.finalDist = finalDist;
	}

	public static Path runAllCmd(double speed,double dist,int cmd){
    	Path path = new Path(speed,dist);
    	DoubleArrayList steps = path.getSteps();
    	DoubleArrayList distance = path.getDist();
    	IntArrayList cmds = path.getCommands();
    	
    	int t=0;
    	while (Math.abs(dist)>waypointIntersectionRadius){    		
    		cmds.add(cmd);
    		speed = updateSpeed(speed, cmd);
    		dist -= speed;
    		steps.add(speed);
    		distance.add(dist);
    		++t;
    	}
    	
    	path.setNoOfSteps(t);
    	path.setMaxSteps(t);
    	path.setFinalDist(dist);
    	path.setFinalSpeed(speed);
    	return path;
    }
	
	public static Path runAllCmd(double speed,double dist,int cmd,int maxSteps){
    	Path path = new Path(speed,dist);
//    	System.out.println(speed);
    	DoubleArrayList steps = path.getSteps();
    	DoubleArrayList distance = path.getDist();
    	IntArrayList cmds = path.getCommands();
    	
    	int t=0;
    	while (Math.abs(dist)>waypointIntersectionRadius && t<=maxSteps){
    		
    		cmds.add(cmd);
    		speed = updateSpeed(speed, cmd);    		
    		dist -= speed;    
    		steps.add(speed);
    		distance.add(dist);
    		++t;
    	}
    	    	
    	path.setNoOfSteps(t);
    	path.setMaxSteps(t);
    	path.setFinalDist(dist);
    	path.setFinalSpeed(speed);
    	return path;
    }
	
	public static Path run(double speed, double dist,IntArrayList cmds,int maxSteps){		
		Path path = new Path(speed,dist);
    	DoubleArrayList steps = path.getSteps();
    	DoubleArrayList distance = path.getDist();    	
    	IntArrayList commds = path.getCommands();
    	
    	int t=0;
    	while (t<cmds.size() && t<maxSteps){
    		int cmd = cmds.getInt(t);    		
    		commds.add(cmd);
    		speed = updateSpeed(speed, cmd);
    		dist -= speed;
    		steps.add(speed);
    		distance.add(dist);
    		++t;
    	}
    	
    	path.setNoOfSteps(t);
    	path.setMaxSteps(maxSteps);
    	path.setFinalDist(dist);
    	path.setFinalSpeed(speed);    	
    	return path;
	}
	
	public static Path combine(Path first,int cut,Path snd){
		if (cut<0 || cut>=first.getNoOfSteps()) return null;
		DoubleArrayList dist = first.getDist();
		DoubleArrayList steps = first.getSteps();
		IntArrayList cmds = first.getCommands();
		
		if (dist.getDouble(cut) != snd.getInitialDist()) return null;
		DoubleArrayList d = new DoubleArrayList(dist.subList(0, cut+1));
		DoubleArrayList s = new DoubleArrayList(steps.subList(0, cut+1));
		IntArrayList c = new IntArrayList(cmds.subList(0, cut+1));
		d.addAll(snd.getDist());
		s.addAll(snd.getSteps());
		c.addAll(snd.getCommands());
		Path rs = new Path(first.initialSpeed,snd.expectedSpeed,first.initialDist,first.maxSteps,d,s,c);
		rs.setNoOfSteps(c.size());
		return rs;
	}
	
	public static Path softBrake(double speed,double dist,double expectedSpeed,int maxSteps){
//		if (speed<expectedSpeed) 
//			return null;		
		Path tmp = Path.runAllCmd(speed, dist, neutral,maxSteps);
		if (tmp.getFinalDist()>waypointIntersectionRadius) {			
			return null;
		}
		if (tmp.getFinalSpeed()<expectedSpeed) return tmp;
		
		Path rs = Path.runAllCmd(speed, dist, backward,maxSteps);
		if (rs.getFinalSpeed()>expectedSpeed) {
//			System.out.println(speed+"   "+dist+"   "+maxSteps);						
			return (rs.getNoOfSteps()==maxSteps && rs.getFinalDist()>waypointIntersectionRadius) ? null : rs;
		}
		Path r=null;
		int i,len;
		IntArrayList cmds = rs.getCommands();
		DoubleArrayList steps = rs.getSteps();
		DoubleArrayList distance = rs.getDist();
		for (i=0,len =rs.getNoOfSteps();i<len;++i){
			double sp = steps.getDouble(i);
			double d = distance.getDouble(i);
			r = Path.runAllCmd(sp, d, neutral, maxSteps-1-i);
			
			if (r.getFinalDist()<=waypointIntersectionRadius && r.getFinalSpeed()<=expectedSpeed)
//				System.out.println(r.getFinalDist()+"    "+r.getFinalSpeed()+"   dfsf");
				break;
		}
//		IntArrayList ar = new IntArrayList(cmds.subList(0, i));
//		for (int j=0;j<i;++j)
//			ar.add(backward);
//		ar.addAll(r.getCommands());
//		return Path.run(speed, dist, ar, maxSteps);
		return combine(rs, i, r);
	}
	
	public static Path run(double speed,double dist,double expectedSpeed,int maxSteps){		
		Path rs = Path.runAllCmd(speed, dist, forward);	
		if (rs.noOfSteps >maxSteps) return null;		
		if (rs.getFinalSpeed()<=expectedSpeed)
			return rs;
		
		if (speed>expectedSpeed ){			
			Path tmp = Path.runAllCmd(speed, dist, backward,maxSteps);
			if (tmp.getFinalDist()<=waypointIntersectionRadius && tmp.getFinalSpeed()>=expectedSpeed)
				return tmp;
		}
		
		int noOfForward = rs.noOfSteps;
		
		Path tmp = rs,val;
		int i;
		IntArrayList cmds = rs.getCommands();
		DoubleArrayList steps = rs.getSteps();
		DoubleArrayList distance = rs.getDist();
		
		for (i=noOfForward-1;i>=0;--i){
			double sp = steps.getDouble(i);
//			if (sp<expectedSpeed) break;
			double d = distance.getDouble(i);
									
			val = Path.softBrake(sp, d, expectedSpeed, maxSteps-i-1);
			
			if (val==null) break;
			tmp = val;		
			if (tmp!=null && tmp.getFinalDist()<=waypointIntersectionRadius && tmp.getFinalSpeed()<=expectedSpeed) break;			
		}
		if (tmp==rs) return rs;
		if (i>=0){
//			System.out.println(tmp.getInitialDist()+"    "+distance.getDouble(i));
//			IntArrayList ar = new IntArrayList(cmds.subList(0, i+1));
//	//		for (int j=0;j<i+1;++j)
//	//			ar.add(forward);
//			System.out.println(Path.run(speed, dist, ar, maxSteps).getDist());
//			ar.addAll(tmp.getCommands());
//			System.out.println(Path.run(speed, dist, ar, maxSteps).getDist());
//			System.out.println(combine(rs,i,tmp).getDist());
			
//			return Path.run(speed, dist, ar, maxSteps);
			return combine(rs,i,tmp);
			
		}
				
		return rs;
	}
	
}
