package drifter;

import simplerace.*;
import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.objects.ObjectArrayList;

public class Turning implements Constants {
	Vehicle car;
	Vector2d wp;
	int maxSteps=1000;
	int noOfSteps;
	DoubleArrayList dist = new DoubleArrayList();
	ObjectArrayList<Vehicle> steps = new ObjectArrayList<Vehicle>();
	IntArrayList commands =  new IntArrayList();
	final static double minSpeedToHardTurn=2.5;
	   
	public Turning() {
		// TODO Auto-generated constructor stub		
	}
    
    public static double getSpeedToNextWaypoint(Vector2d velocity, double orientation)
    {
    	double velocityMagnitude = velocity.mag ();
        double directionOfMovement = Math.atan2 (velocity.y, velocity.x);
        // rotate
        double directionDifference = directionOfMovement - orientation;        
        return Math.cos (directionDifference) * velocityMagnitude;
    }
	
	public Turning(Vehicle car, Vector2d wp, int maxSteps) {
		super();
		this.car = car;
		this.wp = wp;
		this.maxSteps = maxSteps;
	}



	public Turning(Vehicle car, Vector2d wp) {
		super();
		this.car = car;
		this.wp = wp;
	}



	public Vehicle getCar() {
		return car;
	}

	public void setCar(Vehicle car) {
		this.car = car;
	}

	public Vector2d getWp() {
		return wp;
	}

	public void setWp(Vector2d wp) {
		this.wp = wp;
	}

	public int getMaxSteps() {
		return maxSteps;
	}

	public void setMaxSteps(int maxSteps) {
		this.maxSteps = maxSteps;
	}

	public int getNoOfSteps() {
		return noOfSteps;
	}

	public void setNoOfSteps(int noOfSteps) {
		this.noOfSteps = noOfSteps;
	}

	public DoubleArrayList getDist() {
		return dist;
	}

	public void setDist(DoubleArrayList dist) {
		this.dist = dist;
	}

	public ObjectArrayList<Vehicle> getSteps() {
		return steps;
	}

	public void setSteps(ObjectArrayList<Vehicle> steps) {
		this.steps = steps;
	}

	public IntArrayList getCommands() {
		return commands;
	}

	public void setCommands(IntArrayList commands) {
		this.commands = commands;
	}
	
	public static double getAngleToWaypoint(Vector2d pos, double orientation, Vector2d wp)
    {        
        double xDiff = wp.x - pos.x;
        double yDiff = wp.y - pos.y;
        double angle = Math.atan2(yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle(angle);
    }
    
	public static double correctAngle (double angle) {
        while (angle < - Math.PI)
            angle += Math.PI * 2;
        while (angle > Math.PI)
            angle -= Math.PI * 2;
        return angle;
    }
	
	public static Turning turn(Vehicle car,Vector2d wp, Vector2d nwp, int cmd, int maxSteps){
		Vehicle tmp = car.copy();
		Vector2d v = tmp.getVelocity();
		double orientation = tmp.getOrientation();
		double distance = car.getPosition() .dist(wp);
		double theta =  Math.atan((waypointIntersectionRadius)/distance)/2;				
		double heading = (cmd==forwardright || cmd==forwardleft || cmd==left || cmd==right) ?correctAngle(getAngleToWaypoint(tmp.getPosition(), orientation, wp))
			:  correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));
		
		Turning t = new Turning(tmp,wp,maxSteps);
		ObjectArrayList<Vehicle> steps = t.getSteps();
		DoubleArrayList dist = t.getDist();
		IntArrayList cmds = t.getCommands();		
		double velocityMagnitude = v.mag ();
        double directionOfMovement = Math.atan2 (v.y, v.x);
        // rotate
        double directionDifference = directionOfMovement - orientation;
        double speed = Math.cos (directionDifference) * velocityMagnitude;
		int i = 0;
		int cm;
		while (Math.abs(heading)>theta && distance>waypointIntersectionRadius && i<maxSteps){
//			System.out.println(speed);
			cm = cmd;
			if (Math.abs(speed)<minSpeedToHardTurn){
				if (cm==left) cm=forwardleft;
				if (cm==right) cm=forwardright;
			}
			cmds.add(cm);
			
			tmp.update(cm);
			++i;
			v = tmp.getVelocity();
			orientation = tmp.getOrientation();
			distance = car.getPosition() .dist(wp);
			theta = Math.atan((waypointIntersectionRadius)/distance)/2;
			heading = (cm==forwardright || cm==forwardleft || cm==left || cm==right) ?getAngleToWaypoint(tmp.getPosition(), orientation, wp)
					: correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));
			
//			velocityMagnitude = v.mag ();
//	        directionOfMovement = Math.atan2 (v.y, v.x);
//	        // rotate
//	        directionDifference = directionOfMovement - orientation;
//	        speed = Math.cos (directionDifference) * velocityMagnitude;
			dist.add(distance);
			steps.add(tmp.copy());
		
		}
		
//		if (i==maxSteps && distance>waypointIntersectionRadius) return null;
		t.setNoOfSteps(i);
		return t;
	}
		
	
	public static Turning mustHit(Vehicle car,Vector2d wp, Vector2d nwp,double expectedSpeed,int maxSteps){
		if (car==null) return null;
//		if (car.getPosition().dist(new Vector2d(114.98557168635988,304.35049032701244))==0)
//			System.out.println(12345);
//		System.out.println(car.getPosition());
		Turning tmp = forwardOnly(car, wp, nwp,expectedSpeed, maxSteps,false);		
		
		if (tmp == null || tmp.getNoOfSteps()>maxSteps){			
			double orientation = car.getOrientation();
    		double heading = correctAngle(getAngleToWaypoint(car.getPosition(), orientation, wp));
    		int cmd = (heading>0)?right :left;
    		IntArrayList cmds = new IntArrayList();
    		for (int i=0;i<30;++i)
    			cmds.add(cmd);
    		
    		Turning t = runWithAdjust(car, wp, cmds, maxSteps);
    		return tryAll(t, wp, nwp, expectedSpeed, maxSteps);    		
		}		
		
		if (tmp.getNoOfSteps()==0 || tmp.getCar().getPosition().dist(wp)<=waypointIntersectionRadius){				
			return tmp;
		}
				
    	return tryAll(tmp, wp, wp, expectedSpeed, maxSteps);    	
	}
	
	public static Turning combine(Turning first,int cut,Path snd,Vector2d wp){
		if (cut<0 || cut>first.getNoOfSteps()) return null;
		DoubleArrayList dist = first.getDist();
		ObjectArrayList<Vehicle> steps = first.getSteps();
		IntArrayList cmds = first.getCommands();		
//		if (dist.getDouble(cut) != snd.getInitialDist()) return null;
		DoubleArrayList d = new DoubleArrayList(dist.subList(0, cut+1));
		ObjectArrayList<Vehicle> s = new ObjectArrayList<Vehicle>(steps.subList(0, cut+1));
		IntArrayList c = new IntArrayList(cmds.subList(0, cut+1));
		
		Turning t = Turning.run(s.get(s.size()-1), wp, snd.commands, snd.maxSteps);
		
		if (t!=null) {
			
			s.addAll(t.getSteps());
			d.addAll(t.getDist());
			c.addAll(t.getCommands());			
		}
		
		Turning rs = new Turning(t.getCar(),wp,first.maxSteps);
		rs.setCommands(c);
		rs.setDist(d);
		rs.setNoOfSteps(c.size());
		rs.setSteps(s);
		
		return rs;
	}
	
	public static Turning combine(Turning first,int cut,Turning snd,Vector2d wp){
		if (cut<0 || cut>first.getNoOfSteps()) return null;
		DoubleArrayList dist = first.getDist();
		ObjectArrayList<Vehicle> steps = first.getSteps();
		IntArrayList cmds = first.getCommands();
		
//		if (steps.get(cut).getPosition().dist(snd.getCar().getPosition()) !=0) return null;
		DoubleArrayList d = new DoubleArrayList(dist.subList(0, cut+1));
		ObjectArrayList<Vehicle> s = new ObjectArrayList<Vehicle>(steps.subList(0, cut+1));
		IntArrayList c = new IntArrayList(cmds.subList(0, cut+1));
		d.addAll(snd.getDist());
		s.addAll(snd.getSteps());
		c.addAll(snd.getCommands());
		
		Turning rs = new Turning(s.get(s.size()-1),wp,first.maxSteps);
		rs.setCommands(c);
		rs.setDist(d);
		rs.setNoOfSteps(c.size());
		rs.setSteps(s);		
		return rs;
	}
	
	public static boolean checkCondition(double heading,double nextHeading,double theta,double distance,double speed){
//		System.out.println(heading+"    "+nextHeading+"   "+theta);
		if (distance<77 || nextHeading==heading){
			return Math.abs(heading)>theta/2;
		}		
		return (heading*nextHeading<=0 || Math.abs(heading)>=Math.abs(nextHeading) || Math.abs(heading)>theta);
		
	}
	
	public static Turning turnForwardOnly(Vehicle car,Vector2d wp, Vector2d nwp,int maxSteps,boolean strict){
		if (car==null) return null;
		Vehicle tmp = car.copy();
		Vector2d v = tmp.getVelocity();
				
		double orientation = tmp.getOrientation();		
		
		double distance = tmp.getPosition() .dist(wp);
		double theta =  Math.abs(Math.atan((waypointIntersectionRadius)/distance));
		if (strict) theta/=2;
		double heading = correctAngle(getAngleToWaypoint(tmp.getPosition(), orientation, wp));
		double nextHeading = getAngleToWaypoint(tmp.getPosition(), orientation, nwp);		
		double velocityMagnitude = v.mag ();
        double directionOfMovement = Math.atan2 (v.y, v.x);
        // rotate
        double directionDifference = directionOfMovement - orientation;
        double speed = Math.cos (directionDifference) * velocityMagnitude;
//		boolean ok = (heading*nextHeading>0 && Math.abs(nextHeading)>Math.abs(heading));
//		double sign = (heading<0) ? -1 : 1;
		
		 
		Turning t = new Turning(tmp,wp,maxSteps);
		ObjectArrayList<Vehicle> steps = t.getSteps();
		DoubleArrayList dist = t.getDist();
		IntArrayList cmds = t.getCommands();		
		int i = 0;
		int cmd =0;
//		System.out.println(wp+"   "+nwp+"     "+heading+"    "+nextHeading+"    "+theta);
		
//		boolean ok = v.dist(new Vector2d(-6.130760991518019 ,0))==0;
		while (checkCondition(heading, nextHeading,theta,distance,speed) && distance>waypointIntersectionRadius && i<maxSteps){
//			System.out.println(distance);
			cmd = (heading>=Math.PI/2) ? left : (heading>0) ? forwardleft : (heading==0) ? forward : (heading>=-Math.PI/2) ? forwardright : right;
//			if (!ok && heading*sign<0)
//				cmd=(cmd==forwardleft)?forwardright:forwardleft;
			
			if (Math.abs(speed)<=minSpeedToHardTurn){
				if (cmd==left) cmd=forwardleft;
				if (cmd==right) cmd=forwardright;
			}
						
			cmds.add(cmd);						
			tmp.update(cmd);
			++i;
			v = tmp.getVelocity();
			
			orientation = tmp.getOrientation();
			distance = tmp.getPosition() .dist(wp);
			theta = Math.abs(Math.atan((waypointIntersectionRadius)/distance));
			if (strict) theta/=2;
			heading = correctAngle(getAngleToWaypoint(tmp.getPosition(), orientation, wp));
//			nextHeading = getAngleToWaypoint(tmp.getPosition(), orientation, nwp);
			velocityMagnitude = v.mag ();
	        directionOfMovement = Math.atan2 (v.y, v.x);
	        // rotate
	        directionDifference = directionOfMovement - orientation;
	        speed = Math.cos (directionDifference) * velocityMagnitude;	     
			dist.add(distance);
			steps.add(tmp.copy());
		}		
//		System.out.println(i);	
//		if (i==maxSteps && checkCondition(ok, sign, heading, theta) && distance>waypointIntersectionRadius) {
//			System.out.println("here");
//			return null;
//		}
		t.setNoOfSteps(i);
		
		return t;
	}
	
	public static Turning turnBackwardOnly(Vehicle car,Vector2d wp, int maxSteps,boolean strict){
		Vehicle tmp = car.copy();
		Vector2d v = tmp.getVelocity();
		double orientation = tmp.getOrientation();
		
		double distance = tmp.getPosition() .dist(wp);		
		double theta =  Math.atan((waypointIntersectionRadius)/distance)/2;	
		if (strict) theta/=5;
		double heading = correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));		
		Turning t = new Turning(tmp,wp,maxSteps);
		ObjectArrayList<Vehicle> steps = t.getSteps();
		DoubleArrayList dist = t.getDist();
		IntArrayList cmds = t.getCommands();		
		int i = 0;
		int cmd =0;
//		boolean ok = v.dist(new Vector2d(-6.130760991518019 ,0))==0;
		
		while (Math.abs(heading)>theta && i<maxSteps && distance>waypointIntersectionRadius){
			cmd = (heading>0) ? backwardleft : backwardright;
			cmds.add(cmd);
						
			tmp.update(cmd);
			++i;
			v = tmp.getVelocity();
			orientation = tmp.getOrientation();
			distance = tmp.getPosition() .dist(wp);
			theta = Math.atan((waypointIntersectionRadius)/distance)/2;
			if (strict) theta/=5;
			heading = correctAngle(Math.PI - getAngleToWaypoint(tmp.getPosition(), orientation, wp));
			dist.add(distance);
			steps.add(tmp.copy());
		}
		
		
//		if (i==maxSteps && Math.abs(heading)>theta/2 && distance>waypointIntersectionRadius) return null;
//		if (ok) System.out.println(i);		
		t.setNoOfSteps(i);
		return t;
	}
	

	
	public static Turning run(Vehicle car, Vector2d wp,IntArrayList cmds,int maxSteps){
		Vehicle tmp = car.copy();
		Turning path = new Turning(tmp,wp,maxSteps);		
    	ObjectArrayList<Vehicle> steps = path.getSteps();
    	DoubleArrayList distance = path.getDist();    	
    	IntArrayList commds = path.getCommands();
    	double dist = tmp.getPosition().dist(wp);
    	
    	int t=0;
    	
    	while (dist>waypointIntersectionRadius && t<cmds.size() && t<=maxSteps){    	
    		
    		int cmd = cmds.getInt(t);
    		tmp.update(cmd);
    		commds.add(cmd);
    		steps.add(tmp.copy());
    		dist =  tmp.getPosition().dist(wp);
    		distance.add(dist);
    		++t;
    	}
    	
//    	System.out.println(distance);
    	path.setNoOfSteps(t);
    	path.setMaxSteps(maxSteps);    	    	
    	return path;
	}
	
	public static Turning runWithAdjust(Vehicle car, Vector2d wp,IntArrayList cmds,int maxSteps){
		Vehicle tmp = car.copy();
		Turning path = new Turning(tmp,wp,maxSteps);		
    	ObjectArrayList<Vehicle> steps = path.getSteps();
    	DoubleArrayList distance = path.getDist();    	
    	IntArrayList commds = path.getCommands();
    	double dist = tmp.getPosition().dist(wp);
    	double speed = getSpeed(car);
    	int t=0;
    	
    	while (dist>waypointIntersectionRadius && t<cmds.size() && t<=maxSteps){    	    		
    		int cmd = cmds.getInt(t);
    		if (Math.abs(speed)<minSpeedToHardTurn){
				if (cmd==left) cmd=forwardleft;
				if (cmd==right) cmd=forwardright;
			}
    		tmp.update(cmd);
    		commds.add(cmd);
    		steps.add(tmp.copy());
    		dist =  tmp.getPosition().dist(wp);
    		distance.add(dist);
    		speed = getSpeed(tmp);
    		++t;
    	}
    	
//    	System.out.println(distance);
    	path.setNoOfSteps(t);
    	path.setMaxSteps(maxSteps);    	    	
    	return path;
	}
	
	
	public static Turning tryAll(Turning original,Vector2d wp,Vector2d nwp,double expectedSpeed,int maxSteps){
		if (original==null) return null;
		ObjectArrayList<Vehicle> steps =original.getSteps();
		if (isHitTarget(steps, wp)!=null) return original;
		int min =10000;
		Turning rs = null;
		if (!isTheSame(wp, nwp)){			
			for (int i =0,len=steps.size();i<len;++i){
				Vehicle step = steps.get(i);
	    		Turning t = forwardOnly(step, nwp, nwp, expectedSpeed, 500,false);			
//				Turning t = mustHit(step, nwp, nwp, expectedSpeed, 500);
				if (t==null) continue;
	    		Vehicle v = isHitTarget(t.getSteps(), wp);
				if (t!=null && v!=null && min>t.getNoOfSteps()+i+1 && getSpeed(v)<=expectedSpeed){
	    			min = t.getNoOfSteps()+i+1;
//	    			System.out.println(min);
	    			rs = combine(original, i, t, wp);	    			
	    		}
	    		t = backwardOnly(step, nwp, nwp, expectedSpeed, 500,false);	    		
	    		if (t!=null && isHitTarget(t.getSteps(), wp)!=null && min>t.getNoOfSteps()+i+1){
	    			min = t.getNoOfSteps()+i+1;
	    			rs = combine(original, i, t, wp);	    			
	    		}
	    	};
		} else {
			for (int len=steps.size(),i =len-1;i>=0;--i){
				Vehicle step = steps.get(i);
	    		Turning t = forwardOnly(step, wp, nwp, expectedSpeed, maxSteps-i-1,false);
	    		if (t==null) continue;
	    		double speed = getSpeed(t.getCar());
//	    		if (Math.abs(speed)>expectedSpeed && rs!=null) break;
//	    		rs = new IntArrayList(cmds.subList(0, count++));
	    		if (t!=null && getDistance2Wp(t.getCar(), wp)<=waypointIntersectionRadius && min>t.getNoOfSteps()+i+1 && Math.abs(speed) <=expectedSpeed+2){
	    			min = t.getNoOfSteps()+i+1;
	    			rs = combine(original, i, t, wp);	    			
	    		}
	    		t = backwardOnly(step, wp, nwp, expectedSpeed, maxSteps,false);
//	    		rs = new IntArrayList(cmds.subList(0, count++));
	    		if (t!=null && getDistance2Wp(t.getCar(), wp)<=waypointIntersectionRadius && min>t.getNoOfSteps()+i+1){
	    			min = t.getNoOfSteps()+i+1;
	    			rs = combine(original, i, t, wp);	    			
	    		}
	    	};
			
		}
		return rs;
	}
	
	public static Turning adjustForward(Vehicle car,Vector2d wp,Vector2d nwp,Turning t,double expectedSpeed,int maxSteps){
		if (t==null) return null;
    	
       	    	
    	Turning rs = null;
    	if (!isTheSame(wp, nwp)){
    		rs = tryAll(t, wp, nwp, expectedSpeed, 1000);    
//    		System.out.println(rs.getCommands());
    		if (rs!=null && rs.getNoOfSteps()<1000) return rs;
    		
    		rs = tryAll(t, wp, wp, expectedSpeed, 1000);
//    		System.out.println(rs.getCommands());
    		if (rs!=null && rs.getNoOfSteps()<1000) return rs;
    	}
    	if (getDistance2Wp(t.getCar(), wp)<=waypointDrawingRadius)
    		return t;
    	return mustHit(car, wp, wp, expectedSpeed, maxSteps);

    }
		
    
    public static Turning forwardOnly(Vehicle car, Vector2d wp,Vector2d nwp,double expectedSpeed,int maxSteps,boolean strict){
    	if (car==null) return null;
    	Turning t = Turning.turnForwardOnly(car, wp, nwp,  maxSteps,strict);    	
    	if (t==null) return null;
    	
    	Vehicle tmp = t.getCar();    	    	
    	double distance = tmp.getPosition().dist(wp);
    	if (distance<=waypointIntersectionRadius) return t;
//    	IntArrayList cmds = t.getCommands();    	
//    	
    	Vector2d vel = tmp.getVelocity();
    	double orientation = tmp.getOrientation();    	
    	double speed = getSpeedToNextWaypoint(vel, orientation);
//    	Turning path = Turning.goStraight(tmp, wp, nwp, expectedSpeed, maxSteps);
    	Path path = Path.run(speed, distance, expectedSpeed, maxSteps);
    	if (path==null) return null;
//    	cmds.addAll(path.getCommands());    	
//    	return Turning.run(car, wp, cmds, maxSteps);
    	if (t.getNoOfSteps()==0) return Turning.run(car, wp, path.getCommands(), maxSteps);
    	return Turning.combine(t, t.getNoOfSteps()-1, path, wp);
    }
    
    public static Turning backwardOnly(Vehicle car, Vector2d wp,Vector2d nwp,double expectedSpeed,int maxSteps,boolean strict){
    	Turning t = Turning.turnBackwardOnly(car, wp, maxSteps,strict);
    	if (t==null) return null;    	    	
    	
    	Vehicle tmp = t.getCar();
    	double distance = tmp.getPosition().dist(wp);
    	if (distance<=waypointIntersectionRadius) return t;
    	IntArrayList cmds = t.getCommands();    	
    	
    	
    	Vector2d vel = tmp.getVelocity();
    	double orientation = tmp.getOrientation();
    	double speed = getSpeedToNextWaypoint(vel, orientation);
    	
    	Path path = Path.runAllCmd(speed, -distance,backward, maxSteps-t.getNoOfSteps());
//    	System.out.println(path.getCommands());
//    	if (ok) System.out.println(path.getNoOfSteps());
//    	cmds.addAll(path.getCommands());    	
//    	return Turning.run(car, wp, cmds, maxSteps);
    	if (t.getNoOfSteps()==0) return Turning.run(car, wp, path.getCommands(), maxSteps);    	
    	return Turning.combine(t, t.getNoOfSteps()-1, path, wp);
    }

    public static Vehicle isHitTarget(ObjectArrayList<Vehicle> steps,Vector2d wp){
    	for (Vehicle step:steps){
    		if (step.getPosition().dist(wp)<=waypointIntersectionRadius){
//    			System.out.println(step.getPosition().dist(wp));
    			return step;
    		}
    	}
    	return null;
    }
    public static boolean isTheSame(Vector2d v1,Vector2d v2){
    	return (v1.x==v2.x && v1.y==v2.y);
    }
        
    public static IntArrayList adjustBackward(Vehicle car,Vector2d wp,Vector2d nwp,IntArrayList cmds,double expectedSpeed,int maxSteps){
    	Turning t = Turning.run(car, wp, cmds, maxSteps); 
       	ObjectArrayList<Vehicle> steps =t.getSteps();       	
       	if (steps.size()==0) return cmds;
       	
//       	return cmds;
    	Vehicle tmp = steps.get(steps.size()-1);    	
    	boolean hit = (tmp.getPosition().dist(wp)<=waypointIntersectionRadius);
    	
    	int len=cmds.size();
    	
    	IntArrayList rs;    	
    	int min=10000;
    	IntArrayList minCmds=null;
    	int count =0;
    	if (!isTheSame(wp, nwp)){
	    	for (Vehicle step :steps){    		
	    		t = better(forwardOnly(step, nwp, nwp, expectedSpeed, 1000,true),backwardOnly(step, nwp, nwp, expectedSpeed, 1000,true));
	    		rs = new IntArrayList(cmds.subList(0, count++));
	    		if (t!=null && isHitTarget(t.getSteps(), wp)!=null && min>t.getNoOfSteps()+count){
	    			min = t.getNoOfSteps()+count;
	    			rs.addAll(t.getCommands());
	    			minCmds =rs;	    			
	    		}
	    	};
	    	
    	}//end of if

    	boolean allTurn=true;
    	int cmd0 = cmds.getInt(0);
    	if (minCmds==null) {
    		min=len;
        	count =0;
        	
        	
        	for (Vehicle step :steps){
        		int cmd = cmds.getInt(count);
        		if ((cmd==backwardleft) && (cmd0==backwardright)){
        			count++;
        			continue;
        		}
        		allTurn = false;
        		t = backwardOnly(step, wp, nwp, expectedSpeed, maxSteps-count,true);        		
        		rs = new IntArrayList(cmds.subList(0, count++));
        		if (t!=null && isHitTarget(t.getSteps(), wp)!=null && min>t.getNoOfSteps()+count){
        			min = t.getNoOfSteps()+count;
        			rs.addAll(t.getCommands());
        			minCmds =rs;
        		}
        	};
        	if (minCmds==null && hit) return minCmds=cmds;
    	}; 
    	
    	
    	if (minCmds==null){
    		if (!allTurn){
    			min=1000;
    			double heading =Utils.correctAngle(getAngleToWaypoint(car.getPosition(), car.getOrientation(), wp));
    			cmd0 = (heading>0) ? left :right;
    			t = Turning.turn(car, wp, nwp, cmd0, maxSteps);
    			if (t!=null && isHitTarget(t.getSteps(), wp)!=null){
    				minCmds=t.getCommands();
    				min = t.getNoOfSteps();
    				return minCmds;
    				
    			}
    		};
    	}
    	

    	return minCmds;

    }
       
  
    public static Turning better(Turning p1,Turning p2){
    	if (p1==null && p2==null) return null;
    	if (p1==null) return p2;
    	if (p2==null) return p1;
    	return (p1.getNoOfSteps()<p2.getNoOfSteps()) ? p1 :p2;
    }
    
    public static double getSpeed(Vehicle car){
    	return getSpeedToNextWaypoint(car.getVelocity(), car.getOrientation());
    }
    
    public static double getDistance2Wp(Vehicle car,Vector2d wp){
    	return car.getPosition().dist(wp);
    }
    
    public static Turning softBrake(Vehicle car,Vector2d wp,Vector2d nwp,double expectedSpeed,int maxSteps){
//		if (speed<expectedSpeed) 
//			return null;		
		Turning tmp = Turning.turn(car, wp, nwp, neutral, maxSteps);
		Vehicle c = tmp.getCar();
		if (getDistance2Wp(c, wp)>waypointIntersectionRadius) {			
			return null;
		}
		if (getSpeed(c)<expectedSpeed) return tmp;
		
		Turning rs = Turning.turn(car, wp, nwp, backward, maxSteps);
		Vehicle crs = rs.getCar();
		if (getSpeed(crs)>expectedSpeed) {
//			System.out.println(speed+"   "+dist+"   "+maxSteps);						
			return (rs.getNoOfSteps()==maxSteps && getDistance2Wp(crs, wp)>waypointIntersectionRadius) ? null : rs;
		}
		Turning r=null;
		int i,len;
		IntArrayList cmds = rs.getCommands();
		ObjectArrayList<Vehicle> steps = rs.getSteps();
		DoubleArrayList distance = rs.getDist();
		for (i=0,len =rs.getNoOfSteps();i<len;++i){			
			r = Turning.turn(steps.get(i),wp,nwp, neutral, maxSteps-1-i);
			Vehicle step = r.getCar();
			if (getDistance2Wp(step, wp)<=waypointIntersectionRadius && getSpeed(step)<=expectedSpeed)
				break;
		}
		return combine(rs, i, r,wp);
	}
    
    public static Turning goStraight(Vehicle car,Vector2d wp,Vector2d nwp,double expectedSpeed,int maxSteps){		
		Turning rs = Turning.turn(car, wp, nwp, forward, maxSteps);	
		if (rs.noOfSteps >maxSteps) return null;
		Vehicle c = rs.getCar();
		
		if (getSpeed(c)<=expectedSpeed)
			return rs;
		double speed = getSpeed(car);
		double dist = getDistance2Wp(car, wp);
		
		if (speed>expectedSpeed ){			
			Turning tmp = Turning.turn(car, wp, nwp, backward, maxSteps);
			Vehicle cc = tmp.getCar();
			if (getDistance2Wp(cc, wp)<=waypointIntersectionRadius && getSpeed(cc)>=expectedSpeed)
				return tmp;
		}
		
		int noOfForward = rs.noOfSteps;
		
		Turning val=null;
		int i;
		IntArrayList cmds = rs.getCommands();
		ObjectArrayList<Vehicle> steps = rs.getSteps();
		DoubleArrayList distance = rs.getDist();
		
		for (i=noOfForward-1;i>=0;--i){
			val = Turning.softBrake(steps.get(i), wp,nwp,expectedSpeed, maxSteps-i-1);
			
			if (val==null) break;			
			Vehicle cc = val.getCar();
			if (val!=null && getDistance2Wp(cc, wp)<=waypointIntersectionRadius && getSpeed(cc)<=expectedSpeed) break;			
		}
		if (val==null) return rs;
		if (i>=0){
			return combine(rs,i,val,wp);			
		}
		return rs;
    }

}
